from qtrf.widgets.meters          import prefix_map, units
from qtrf.widgets.value.line_edit import ValueLineEdit

class MetersLineEdit(ValueLineEdit):
    def __init__(self, parent=None):
        ValueLineEdit.__init__(self, parent)
        self.prefix_map       = prefix_map
        self.include_negative = False
        self.units            = units
        self.update_validation()



from PySide2.QtCore    import QRegExp
from PySide2.QtGui     import QRegExpValidator
from PySide2.QtWidgets import QLineEdit

# hostname regex
# https://stackoverflow.com/a/106223/1390788
hostname   = r'(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])'


# ip address regex
from_zero  = r'(((00)?\d)|([01]?\d\d)|(2[0-4]\d)|(25[0-5]))'
from_one   = r'(((0|00)?[1-9])|(0?[1-9]\d)|(1\d\d)|(2[0-4]\d)|(25[0-5]))'
dot        = r'\.'
ip_address = '{from_one}{dot}{from_zero}{dot}{from_zero}{dot}{from_one}'.format(from_zero=from_zero, from_one=from_one, dot=dot)

# total regex
regex_str = '({hostname})|({ip_address})'.format(hostname=hostname, ip_address=ip_address)


class IPAddressLineEdit(QLineEdit):
    def __init__(self, parent=None):
        QLineEdit.__init__(self, parent)
        validator = QRegExpValidator(QRegExp(regex_str))
        self.setValidator(validator)

    def focusInEvent(self, event):
        self.selectAll()



port_regex          = r'[1-9]\d*'
range_regex         = f'\\-{port_regex}'
port_or_range_regex = f'{port_regex}({range_regex})?'
is_valid_regex      = f'^{port_or_range_regex}(, ?{port_or_range_regex})*$'

def entry_to_list(entry):
    list = []
    for item in filter(None, entry.split(',')):
        new_list = _entry_item_to_list(item)
        for i in new_list:
            if not i in list:
                list.append(i)
    list.sort()
    return list
def _entry_item_to_list(item):
    if '-' in item:
        start, stop = map(int, item.split('-'))
        return list(range(start, stop + 1))
    # else
    return [int(item)]
def list_to_entry(list):
    list = list.copy()
    entries = []
    while list:
        if len(list) == 1:
            value_str = str(list.pop())
            entries.append(value_str)
            break;
        range_start = list.pop(0)
        for index, value in enumerate(list):
            if value != range_start + index + 1:
                # passed stop_range
                if index == 0:
                    entries.append(str(range_start))
                else:
                    range_stop = list[index-1]
                    entries.append(f'{range_start}-{range_stop}')
                    list = list[index:]
                break
            elif index == len(list) - 1:
                # current item is end of range
                entries.append(f'{range_start}-{value}')
                list.clear()
                break
    return ', '.join(entries)



from PySide2.QtWidgets import QLabel
from .mixin.list      import ListMixin

class PortsLabel(ListMixin, QLabel):
    def __init__(self, parent=None):
        QLabel   .__init__(self, parent)
        ListMixin.__init__(self)



from PySide2.QtCore    import QRegExp
from PySide2.QtGui     import QRegExpValidator
from PySide2.QtWidgets import QLineEdit


# regex
POINTS_REGEX = r'[1-9]\d*'


def int_or_none(value):
    if value is None:
        return None
    if value == '':
        return None

    return int(value)


class PointsLineEdit(QLineEdit):
    def __init__(self, parent=None):
        QLineEdit.__init__(self, parent)
        self._set_validator()

    @property
    def value(self):
        if not self.text():
            return None

        return int(self.text())

    @value.setter
    def value(self, value):
        if value == self.value:
            # nothing to do
            return

        # set
        value = int_or_none(value)
        if value is None:
            self.clear()
        else:
            self.setText(str(value))

    def focusInEvent(self, event):
        QLineEdit.focusInEvent(self, event)
        self.selectAll()

    def mousePressEvent(self, event):
        QLineEdit.mousePressEvent(self, event)
        self.selectAll()

    # helpers
    def _set_validator(self):
        validator = QRegExpValidator(QRegExp(POINTS_REGEX))
        self.setValidator(validator)



from PySide2.QtWidgets import QLineEdit
from .mixin.focus     import FocusMixin
from .mixin.list      import ListMixin
from .mixin.validator import ValidatorMixin

class PortsLineEdit(ValidatorMixin, FocusMixin, ListMixin, QLineEdit):
    def __init__(self, parent=None):
        QLineEdit     .__init__(self, parent)
        ListMixin     .__init__(self)
        FocusMixin    .__init__(self)
        ValidatorMixin.__init__(self)



from .ui_file                   import Ui_FileWidget
from pathlib                    import Path
from PySide2.QtCore             import Slot
from PySide2.QtWidgets          import QFileDialog, QWidget

class FileWidget(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.ui = Ui_FileWidget()
        self.ui.setupUi(self)
        self.caption    = 'Open...'
        self.filter     = ''
        self._filename  = None
        self._last_path = ''
        self.ui.open .clicked.connect(self.choose_file)
        self.ui.clear.clicked.connect(self.clear)

    @property
    def filename(self):
        return self._filename
    @filename.setter
    def filename(self, value):
        self._filename  = value or ''
        self._last_path = str(Path(value).parent) if value else self._last_path
        self._update_filename()

    @Slot()
    def choose_file(self, dir=None):
        filename,_    = self._dialog(dir)
        self.filename = filename or self.filename
        return filename

    @Slot()
    def clear(self):
        self.filename = None

    def _dialog(self, dir=None):
        return QFileDialog.getOpenFileName(self.window(),
                                           self.caption,
                                           dir or self._last_path,
                                           self.filter)
    def _update_filename(self):
        if self._filename:
            self.ui.filename.setText(Path(self._filename).name)
        else:
            self.ui.filename.clear()



from ..ports import entry_to_list, list_to_entry

class ListMixin(object):
    @property
    def list(self):
        return entry_to_list(self.text())
    @list.setter
    def list(self, ports):
        self.setText(list_to_entry(ports))



from qtrf.widgets.value.label import ValueLabel

class CelsiusLabel(ValueLabel):
    def __init__(self, parent=None):
        ValueLabel.__init__(self, parent)
        self.decimal_places = 1
        self.prefix_map     = None
        self.units          = '°C'



from qtrf.widgets.meters       import prefix_map, units
from qtrf.widgets.value.label  import ValueLabel

class MetersLabel(ValueLabel):
    def __init__(self, parent=None):
        ValueLabel.__init__(self, parent)
        self.prefix_map = prefix_map
        self.units      = units



from PySide2.QtCore    import Qt
from PySide2.QtWidgets import QLineEdit

class FocusMixin(object):
    def __init__(self):
        self._special_keys = [Qt.Key_Enter, Qt.Key_Return]
    def keyPressEvent(self, event):
        key = event.key()
        if self._is_special_key(key):
            self.clean_intermediate_input()
            self.selectAll()
            event.accept()
        else:
            QLineEdit.keyPressEvent(self, event)
    def _is_special_key(self, key):
        return key in self._special_keys

    def focusInEvent(self, event):
        QLineEdit.focusInEvent(self, event)
        self.selectAll()
    def focusOutEvent(self, event):
        self.clean_intermediate_input()
        QLineEdit.focusOutEvent(self, event)



from ..ports        import is_valid_regex
from PySide2.QtCore  import QRegExp
from PySide2.QtGui   import QRegExpValidator

def clean_intermediate_input(text):
    return text.strip().rstrip(',').rstrip('-')
def create_validator():
    return QRegExpValidator(QRegExp(is_valid_regex))

class ValidatorMixin(object):
    def __init__(self):
        self.update_validation()
    def update_validation(self):
        self.setValidator(create_validator())
    def clean_intermediate_input(self):
        self.setText(clean_intermediate_input(self.text()))
        self.list = self.list



from qtrf.widgets.value.line_edit import ValueLineEdit

class CelsiusLineEdit(ValueLineEdit):
    def __init__(self, parent=None):
        ValueLineEdit.__init__(self, parent)
        self.decimal_places = 1
        self.prefix_map     = None
        self.units          = '°C'
        self.update_validation()



from   ddt        import ddt, data
from   decimal    import Decimal
from   qtrf.units import prefix_map, to_decimal, to_str
import unittest

frequency_prefix_map = prefix_map.copy()
del(frequency_prefix_map['f'])
del(frequency_prefix_map['p'])
del(frequency_prefix_map['n'])
del(frequency_prefix_map['m'])

farads_prefix_map = prefix_map.copy()
del(farads_prefix_map['K'])
del(farads_prefix_map['M'])
del(farads_prefix_map['G'])
del(farads_prefix_map['T'])

@ddt
class TestUnits(unittest.TestCase):
    @data({'text': None,     'units': '',  'value': None            },
          {'text': '',       'units': '',  'value': None            },
          {'text': 'm',      'units': '',  'value': None            },
          {'text': 'M',      'units': '',  'value': None            },
          {'text': '0.',     'units': 'U', 'value': Decimal('0')    },
          {'text': '1.',     'units': 'U', 'value': Decimal('1')    },
          {'text': '0',      'units': 'U', 'value': Decimal('0')    },
          {'text': '1',      'units': 'U', 'value': Decimal('1')    },
          {'text': '1.0',    'units': 'U', 'value': Decimal('1')    },
          {'text': '1 m',    'units': 'U', 'value': Decimal('1e-3') },
          {'text': '1 M',    'units': '',  'value': Decimal('1e6')  },
          {'text': '2.1 GU', 'units': 'U', 'value': Decimal('2.1e9')})
    def test_to_decimal(self, data):
        text  = data['text' ]
        units = data['units']
        value = data['value']
        self.assertEqual(to_decimal(text, units, ignore_prefix=False), value)

    @data({'text': None,     'units': '',  'value': None            },
          {'text': '',       'units': '',  'value': None            },
          {'text': 'm',      'units': '',  'value': None            },
          {'text': 'M',      'units': '',  'value': None            },
          {'text': '0.',     'units': 'U', 'value': Decimal('0')    },
          {'text': '1.',     'units': 'U', 'value': Decimal('1')    },
          {'text': '0',      'units': 'U', 'value': Decimal('0')    },
          {'text': '1',      'units': 'U', 'value': Decimal('1')    },
          {'text': '1.0',    'units': 'U', 'value': Decimal('1')    },
          {'text': '1 m',    'units': 'U', 'value': Decimal('1')    },
          {'text': '1 M',    'units': '',  'value': Decimal('1')    },
          {'text': '2.1 GU', 'units': 'U', 'value': Decimal('2.1')  })
    def test_to_decimal_ignore_prefix(self, data):
        text  = data['text' ]
        units = data['units']
        value = data['value']
        self.assertEqual(to_decimal(text, units, ignore_prefix=True), value)

    @data({'value': Decimal('0'),       'decimals': 3, 'prefix_map': prefix_map,           'units': '',    'text': '0.000'      },
          {'value': Decimal('1'),       'decimals': 3, 'prefix_map': prefix_map,           'units': '',    'text': '1.000'      },
          {'value': Decimal('2'),       'decimals': 1, 'prefix_map': prefix_map,           'units': '',    'text': '2.0'        },
          {'value': Decimal('1.005e3'), 'decimals': 2, 'prefix_map': prefix_map,           'units': '',    'text': '1.01 K'     },
          {'value': Decimal('100.3e3'), 'decimals': 3, 'prefix_map': frequency_prefix_map, 'units': 'Hz',  'text': '100.300 KHz'},
          {'value': Decimal('3.01e-5'), 'decimals': 0, 'prefix_map': farads_prefix_map,    'units': 'F',   'text': '30 uF'      },
          {'value': Decimal('1234.56'), 'decimals': 1, 'prefix_map': None,                 'units': 'dBm', 'text': '1234.6 dBm' },
          {'value': Decimal('-1e-2'),   'decimals': 2, 'prefix_map': None,                 'units': 'dBm', 'text': '-0.01 dBm'  })
    def test_to_str(self, data):
        value      = data['value'     ]
        decimals   = data['decimals'  ]
        prefix_map = data['prefix_map']
        units      = data['units'     ]
        text       = data['text'      ]
        self.assertEqual(to_str(value, decimals, prefix_map, units), text)



from   ddt                      import ddt, data
from   qtrf.widgets.ports.ports import entry_to_list, is_valid_regex, list_to_entry
import re
import unittest

@ddt
class TestUnits(unittest.TestCase):
    @data({'entry': '',           'list': []},
          {'entry': '1',          'list': [1]},
          {'entry': '1,2',        'list': [1,2]},
          {'entry': '1, 2, 3, 4', 'list': [1,2,3,4]},
          {'entry': '1-4',        'list': [1,2,3,4]},
          {'entry': '1, 3, 1-4',  'list': [1,2,3,4]})
    def test_entry_to_list(self, data):
        entry = data['entry']
        list  = data['list' ]
        self.assertEqual(entry_to_list(entry), list)

    @data({'list': [],                'entry': ''    },
          {'list': [1],               'entry': '1'   },
          {'list': [1,2],             'entry': '1-2' },
          {'list': [1,3],             'entry': '1, 3'},
          {'list': [1,4,9,10,11],     'entry': '1, 4, 9-11'},
          {'list': [1,2,3,4,9,10,11], 'entry': '1-4, 9-11' })
    def test_list_to_entry(self, data):
        list  = data['list']
        entry = data['entry']
        self.assertEqual(list_to_entry(list), entry)
    @data({'entry': '',           'is_valid': False},
          {'entry': '1',          'is_valid': True},
          {'entry': '1,2',        'is_valid': True},
          {'entry': '1, 2',       'is_valid': True},
          {'entry': '1-4',        'is_valid': True},
          {'entry': '1-2,3, 4-5', 'is_valid': True})
    def test_is_valid_regex(self, data):
        entry    = data['entry']
        is_valid = data['is_valid']
        self.assertEqual(bool(re.match(is_valid_regex, entry)), is_valid)



from PySide2.QtCore    import Signal
from qtrf.units        import prefix_map, to_decimal, to_str


def float_or_none(value):
    if value is None:
        return None
    if value == '':
        return None

    return float(value)


# Use requires:
# - class be QWidget
# - methods: text() and setText(text)
class ValueMixin(object):
    def __init__(self):
        self.decimal_places = 3
        self.prefix_map     = prefix_map.copy()
        self.units          = 'U'

    # value
    @property
    def value(self):
        return to_decimal(self.text(), self.units)

    @value.setter
    def value(self, value):
        value = float_or_none(value)
        if self.value == value:
            # nothing new to do
            return

        # set
        if value is None:
            self.clear()
        else:
            self.setText(self._to_str(value))

        # signal value changed
        self.value_changed.emit(self.value)

    # signals
    value_changed = Signal(float)

    # helper
    def _to_str(self, value):
        return to_str(value, self.decimal_places, self.prefix_map, self.units)



from qtrf.widgets.amps            import prefix_map, units
from qtrf.widgets.value.line_edit import ValueLineEdit

class AmpsLineEdit(ValueLineEdit):
    def __init__(self, parent=None):
        ValueLineEdit.__init__(self, parent)
        self.prefix_map = prefix_map
        self.units      = units
        self.update_validation()



from qtrf.widgets.value.label import ValueLabel

class dBLabel(ValueLabel):
    def __init__(self, parent=None):
        ValueLabel.__init__(self, parent)
        self.prefix_map = None
        self.units      = 'dB'
